#if defined(_WIN32) || defined(_WIN64)
  #include <windows.h>
  #include <tchar.h>
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <iostream>
using std::cerr;
using std::endl;

#include <algorithm>
using std::min;
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <vector>
using std::vector;

#include "acquirefile_read.h"

#ifdef _WIN64
  #define FSEEK _fseeki64
#else
  #define FSEEK fseek
#endif

inline int64_t round_up(int64_t x, int64_t divisor)
{
  return divisor * ((x/divisor) + ((x%divisor) ? 1 : 0));
}

// struct memory management:

static void acquirefile_read_channel_cleanup(acquirefile_read_channel *channel)
{
  delete [] channel->blocks;
}

static void acquirefile_read_segment_cleanup(acquirefile_read_segment *segment, int64_t analog_in_channel_count)
{
  for (int64_t i=analog_in_channel_count-1; i>=0; --i)
    acquirefile_read_channel_cleanup(segment->channels+i);
  delete [] segment->channels;
}

static acquirefile_read_sweep* acquirefile_read_sweep_alloc(int64_t analog_in_channel_count, int64_t aperiodic_channel_count)
{
  acquirefile_read_sweep* sweep = new acquirefile_read_sweep;
  memset(sweep, 0, sizeof(acquirefile_read_sweep));
  sweep->analog = new AcquireSweepPrefixAnalogChannel[analog_in_channel_count];
  sweep->aperiodic = new AcquireSweepPrefixAperiodicChannel[aperiodic_channel_count];
  return sweep;
}

static void acquirefile_read_sweep_free(acquirefile_read_sweep *sweep, int64_t analog_in_channel_count)
{
  if ( sweep ) {
    for (int64_t i=sweep->descriptor.segment_count-1; i>=0; --i)
      acquirefile_read_segment_cleanup(sweep->segments+i, analog_in_channel_count);
    delete [] sweep->segments;
    delete [] sweep->channel_directory;
    delete [] sweep->segment_directory;
    delete [] sweep->control_channels_ext;
    delete [] sweep->control_channels;
    delete [] sweep->aperiodic;
    delete [] sweep->analog;
    delete sweep;
  }
}

static void acquirefile_read_series_cleanup(acquirefile_read_series& series, int64_t analog_in_channel_count)
{
  delete [] series.sweep_directory;
  for (int64_t i=0; i<series.descriptor.sweep_count; ++i)
    acquirefile_read_sweep_free(series.sweeps[i], analog_in_channel_count);
  delete [] series.sweeps;
}

// Reading:

static size_t acquirefile_read_read_tag(FILE *fp, int64_t offset, AcquireBlockTag *tag)
{
  if ( 0 == FSEEK(fp, offset, SEEK_SET) )
    return fread(tag, sizeof(AcquireBlockTag), 1, fp);
  else
    return 0;
}

static acquirefile_read* acquirefile_read_read(FILE *fp)
{
  size_t count_read;
  AcquireBlockTag tag;
  int64_t offset, i, j, k;
  
  // zero and read header:
  acquirefile_read *af = new acquirefile_read;
  memset(af, 0, sizeof(AcquirePrefix) + sizeof(AcquireHeader));
  af->storage = fp;
  FSEEK(fp, 0L, SEEK_SET);
  count_read = fread(&(af->prefix), sizeof(AcquirePrefix), 1, fp);
  count_read = fread(&(af->header), sizeof(AcquireHeader), 1, fp);
  
  // alloc dynamic arrays
  af->analog_in = new AcquireAnalogInChannelRecord[af->header.analog_in_channel_count];
  memset(af->analog_in, 0, af->header.analog_in_channel_count * sizeof(AcquireAnalogInChannelRecord));
  af->aperiodic = new AcquireAperiodicChannelRecord[af->header.aperiodic_channel_count];
  memset(af->aperiodic, 0, af->header.aperiodic_channel_count * sizeof(AcquireAperiodicChannelRecord));
  af->series = new acquirefile_read_series[af->header.series_count];
  memset(af->series, 0, af->header.series_count * sizeof(acquirefile_read_series));

  acquirefile_read_series *series = af->series; // will read in next series then increment pointer
  vector<acquirefile_read_sweep*> sweeps;   // will copy into series->sweeps when we know sweep_count
  acquirefile_read_sweep *sweep = acquirefile_read_sweep_alloc(af->header.analog_in_channel_count, af->header.aperiodic_channel_count);
  // will read next sweep, add to sweeps, then allocate another
  
  offset = round_up(ftell(fp), ACQUIRE_BLOCK_SIZE); // blocks start at multiples of ACQUIRE_BLOCK_SIZE
  while ( acquirefile_read_read_tag(fp, offset, &tag) ) { // will stop looping after ACQUIRE_CONTROL_BLOCK
    switch ( tag.block_type ) {
    case ACQUIRE_DATA_BLOCK:
      break; // read as needed during _read_data_()
    case ACQUIRE_SWEEP_BLOCK:
      count_read = fread(&(sweep->descriptor), sizeof(AcquireSweepDescriptor), 1, fp);
      count_read = fread(&(sweep->control_channel_count), sizeof(int64_t), 1, fp);
      sweep->control_channels = new AcquireBlockDescriptor[ sweep->control_channel_count ];
      count_read = fread(sweep->control_channels, sizeof(AcquireBlockDescriptor), sweep->control_channel_count, fp);
      count_read = fread(&(sweep->control_channel_ext_count), sizeof(int64_t), 1, fp);
      sweep->control_channels_ext = new AcquireBlockDescriptor[ sweep->control_channel_ext_count ];
      count_read = fread(sweep->control_channels_ext, sizeof(AcquireBlockDescriptor), sweep->control_channel_ext_count, fp);
      sweep->segment_directory = new int64_t[sweep->descriptor.segment_count];
      count_read = fread(sweep->segment_directory, sizeof(int64_t), sweep->descriptor.segment_count, fp);
      sweep->channel_directory = new int64_t[af->header.analog_in_channel_count];
      count_read = fread(sweep->channel_directory, sizeof(int64_t), af->header.analog_in_channel_count, fp);
      sweep->segments = new acquirefile_read_segment[ sweep->descriptor.segment_count ];
      for (i=0; i<sweep->descriptor.segment_count; ++i) {
	count_read = fread(&(sweep->segments[i].descriptor), sizeof(AcquireSegmentDescriptor), 1, fp);
	sweep->segments[i].channels = new acquirefile_read_channel[ af->header.analog_in_channel_count ];
	for (j=0; j<af->header.analog_in_channel_count; ++j) {
	  count_read = fread(&(sweep->segments[i].channels[j].descriptor), sizeof(AcquireChannelDescriptor), 1, fp);
	  sweep->segments[i].channels[j].blocks = new AcquireBlockDescriptor[ sweep->segments[i].channels[j].descriptor.block_count ];
	  count_read = fread(sweep->segments[i].channels[j].blocks, sizeof(AcquireBlockDescriptor),
			     sweep->segments[i].channels[j].descriptor.block_count, fp);
	  if ( j == 0 ) {
		sweep->segments[i].frames = 0;
	    for (k=0; k<sweep->segments[i].channels[j].descriptor.block_count; ++k)
	      sweep->segments[i].frames += sweep->segments[i].channels[j].blocks[k].data_count;
	  }
	}
      }
      sweeps.push_back(sweep);
      sweep = acquirefile_read_sweep_alloc(af->header.analog_in_channel_count, af->header.aperiodic_channel_count);
      break;
    case ACQUIRE_CONTROL_BLOCK:
      memcpy(&af->control_tag, &tag, sizeof(AcquireBlockTag));
      af->control_block = new char[tag.block_size];
      memcpy(af->control_block, &tag, sizeof(AcquireBlockTag));
      count_read = fread(((char*) af->control_block)+sizeof(AcquireBlockTag), tag.block_size-sizeof(AcquireBlockTag), 1, fp);
      goto DONE_FILE;
    case ACQUIRE_SERIES_BLOCK:
      count_read = fread(&(series->descriptor), sizeof(AcquireSeriesDescriptor), 1, fp);
      series->sweep_directory = new int64_t[series->descriptor.sweep_count];
      count_read = fread(series->sweep_directory, sizeof(int64_t), series->descriptor.sweep_count, fp);
      series->sweeps = new acquirefile_read_sweep*[series->descriptor.sweep_count];
      memcpy(series->sweeps, &(sweeps[0]), series->descriptor.sweep_count*sizeof(acquirefile_read_sweep*));
      ++series;
      break;
    case ACQUIRE_FILE_PREFIX_BLOCK:
      for (i=0; i<af->header.analog_in_channel_count; ++i)
	count_read = fread(af->analog_in+i, sizeof(AcquireAnalogInChannelRecord), 1, fp);
      for (i=0; i<af->header.aperiodic_channel_count; ++i)
	count_read = fread(af->aperiodic+i, sizeof(AcquireAperiodicChannelRecord), 1, fp);
      break;
    case ACQUIRE_SWEEP_PREFIX_BLOCK:
      for (i=0; i<af->header.analog_in_channel_count; ++i)
	count_read = fread(sweep->analog+i, sizeof(AcquireSweepPrefixAnalogChannel), 1, fp);
      for (i=0; i<af->header.aperiodic_channel_count; ++i)
	count_read = fread(sweep->aperiodic+i, sizeof(AcquireSweepPrefixAperiodicChannel), 1, fp);
      break;
    }
    offset += tag.block_size;
  }
 DONE_FILE:
  
  // extra/unused
  acquirefile_read_sweep_free(sweep, af->header.analog_in_channel_count);

  // buffers for data reading, decoding, scaling
  af->data_block_buf = new int8_t[af->header.analog_in_maximum_block_size];
  af->data_buf = new int16_t[af->header.analog_in_maximum_data_count];
  
  return af;
}


// Data is stored as compressed, big-endian int16_t.  The "last_value" implicitly starts at 0.
// If the next compressed byte b is (-127 <= b <= 127), then next_value = last_value + b.
// If the next compressed byte b == -128, then the following two bytes are interpreted as
// a big-endian 16-bit integer (i.e. high-order, then low-order byte).

static bool acquirefile_read_decode(acquirefile_read *file, AcquireBlockDescriptor *block)
{
  FSEEK((FILE*) file->storage, block->location, SEEK_SET);
  size_t count_read = fread(file->data_block_buf, block->size, 1, (FILE*) file->storage);
  if ( ! count_read )
    return false;
  
  int8_t *src = file->data_block_buf + sizeof(AcquireBlockTag);
  int8_t *src_end = src + block->data_size;
  int16_t *dest = file->data_buf;
  int16_t *dest_end = dest + file->header.analog_in_maximum_data_count;
  int16_t val = 0;
  int8_t *val_bytes = (int8_t *) &val; // to convert 16-bit literals from big-endian
  while ( (src != src_end) && (dest != dest_end) ) {
    int8_t c = *src;
    if ( c == -128 ) {
      ++src;
      val_bytes[1] = *src;
      ++src;
      val_bytes[0] = *src;
      ++src;
    } else {
      val += c;
      ++src;
    }
    *dest = val;
    ++dest;
  }
  
  return true;
}


extern "C" ACQUIREFILE_API acquirefile_read *acquirefile_read_open(const char *path)
{
  FILE *fp = fopen(path, "rb");
  if ( ! fp ) {
    perror("acquirefile: fopen");
    return NULL;
  }
  return acquirefile_read_read(fp);
}

#if defined(_WIN32) && defined(UNICODE)
extern "C" ACQUIREFILE_API acquirefile_read *acquirefile_read_open_w(const _TCHAR *path)
{
  FILE *fp = _tfopen(path, "rb");
  if ( ! fp ) {
    perror("acquirefile: _tfopen");
    return NULL;
  }
  return acquirefile_read_read(fp);
}
#endif

extern "C" ACQUIREFILE_API void acquirefile_read_close(acquirefile_read *file)
{
  if ( ! file )
    return;

  fclose( (FILE*) file->storage );
  delete [] file->analog_in;
  delete [] file->aperiodic;
  delete [] (char*) file->control_block;
  for (int64_t i=0; i<file->header.series_count; ++i)
    acquirefile_read_series_cleanup(file->series[i], file->header.analog_in_channel_count);
  delete [] file->series;
  delete [] file->data_block_buf;
  delete [] file->data_buf;
  delete file;
}

extern "C" ACQUIREFILE_API int64_t acquirefile_read_data_int16(acquirefile_read *file,
						    int64_t series, int64_t sweep, int64_t segment, int64_t channel,
						    int64_t offset, int64_t count, int16_t *buf)
{
  int64_t at, block_i, into, start;
  acquirefile_read_channel *datachan = & file->series[series].sweeps[sweep]->segments[segment].channels[channel];
  at = 0;
  block_i = 0;
  // find first relevant block
  while ( block_i < datachan->descriptor.block_count ) {
    if ( offset < (at + datachan->blocks[block_i].data_count) ) {
      break;
    } else {
      at += datachan->blocks[block_i].data_count;
      ++block_i;
    }
  }
  into = 0;
  while ( count && (block_i < datachan->descriptor.block_count) ) {
    start = offset - at;
    int64_t count_here = min(count, datachan->blocks[block_i].data_count - start);
    if ( ! acquirefile_read_decode(file, datachan->blocks+block_i) )
      return into;
    memcpy(buf+into, file->data_buf+start, sizeof(int16_t)*count_here);
    into += count_here;
    count -= count_here;
    offset += count_here;
    at += datachan->blocks[block_i].data_count;
    ++block_i;
  }
  return into;
}

extern "C" ACQUIREFILE_API int64_t acquirefile_read_data_float(acquirefile_read *file,
						    int64_t series, int64_t sweep, int64_t segment, int64_t channel,
						    int64_t offset, int64_t count, float *buf)
{
  // the signal is scaled from "values" (int16_t) to volts, then to SI units (e.g. "A"), then to natural units (e.g. "pA")
  AcquireSweepPrefixAnalogChannel *ch_prefix = file->series[series].sweeps[sweep]->analog + channel;
  float scale = (float) ( (1.0 / (ch_prefix->gain * pow(1e3, (double)file->analog_in[channel].natural_scale)))
		  * (ch_prefix->maximum_voltage - ch_prefix->minimum_voltage) * 1.0 / (ch_prefix->maximum_value - ch_prefix->minimum_value) );

  int64_t at, block_i, into, start, i;
  acquirefile_read_channel *datachan = & file->series[series].sweeps[sweep]->segments[segment].channels[channel];
  at = 0;
  block_i = 0;
  // find first relevant block
  while ( block_i < datachan->descriptor.block_count ) {
    if ( offset < (at + datachan->blocks[block_i].data_count) ) {
      break;
    } else {
      at += datachan->blocks[block_i].data_count;
      ++block_i;
    }
  }
  into = 0;
  while ( count && (block_i < datachan->descriptor.block_count) ) {
    start = offset - at;
    int64_t count_here = min(offset+count, at+datachan->blocks[block_i].data_count) - offset;
    if ( ! acquirefile_read_decode(file, datachan->blocks+block_i) )
      return into;
    for (i=0; i<count_here; ++i)
      buf[into+i] = scale * file->data_buf[start+i];
    into += count_here;
    count -= count_here;
    offset += count_here;
    at += datachan->blocks[block_i].data_count;
    ++block_i;
  }
  return into;
}



